@@ -2,6 +2,7 @@
  * ctl.c
  *
  * (C) 2019 by Sylvain Munaut <tnt@246tNt.com>
+ * (C) 2020 by Harald Welte <laforge@gnumonks.org>
  *
  * All Rights Reserved
  *
@@ -25,6 +26,7 @@
 #include <fcntl.h>
 #include <unistd.h>
 #include <string.h>
+#include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/socket.h>
@@ -40,34 +42,15 @@
 #include "log.h"
 
 
-struct e1_intf *
-e1d_find_intf(struct e1_daemon *e1d, uint8_t id)
-{
-	struct e1_intf *intf;
-
-	llist_for_each_entry(intf, &e1d->interfaces, list)
-		if (intf->id == id)
-			return intf;
-
-	return NULL;
-}
-
-struct e1_line *
-e1_intf_find_line(struct e1_intf *intf, uint8_t id)
-{
-	struct e1_line *line;
-
-	llist_for_each_entry(line, &intf->lines, list)
-		if (line->id == id)
-			return line;
-
-	return NULL;
-}
-
 static struct e1_ts *
 _e1d_get_ts(struct e1_line *line, uint8_t ts)
 {
-	return (ts < 32) ? &line->ts[ts] : NULL;
+	if (ts < 32)
+		return &line->ts[ts];
+	else if (ts == E1DP_TS_SUPERCHAN)
+		return &line->superchan;
+	else
+		return NULL;
 }
 
 static void
@@ -81,6 +64,16 @@ static void
 _e1d_fill_line_info(struct osmo_e1dp_line_info *li, struct e1_line *line)
 {
 	li->id = line->id;
+	switch (line->mode) {
+	case E1_LINE_MODE_CHANNELIZED:
+		li->cfg.mode = E1DP_LMODE_CHANNELIZED;
+		break;
+	case E1_LINE_MODE_SUPERCHANNEL:
+		li->cfg.mode = E1DP_LMODE_SUPERCHANNEL;
+		break;
+	default:
+		OSMO_ASSERT(0);
+	}
 	li->status = 0x00;
 }
 
@@ -104,16 +97,43 @@ e1_ts_stop(struct e1_ts *ts)
 		close(ts->fd);
 		ts->fd = -1;
 	}
+
+	talloc_free(ts->raw.rx_buf);
+	ts->raw.rx_buf = NULL;
+	ts->raw.rx_buf_size = 0;
+	ts->raw.rx_buf_used = 0;
+}
+
+static void
+_e1d_ts_raw_buf_realloc(struct e1_ts *ts, unsigned int size)
+{
+	ts->raw.rx_buf = talloc_realloc_size(ts->line, ts->raw.rx_buf, size);
+	OSMO_ASSERT(ts->raw.rx_buf);
+	ts->raw.rx_buf_size = size;
+	ts->raw.rx_buf_used = 0;
 }
 
 static int
-_e1d_ts_start(struct e1_ts *ts, enum e1_ts_mode mode)
+_e1d_ts_start(struct e1_ts *ts, enum e1_ts_mode mode, uint16_t bufsize)
 {
 	int ret, sd[2];
+	int sock_type;
 
-	LOGPTS(ts, DE1D, LOGL_INFO, "Starting in mode %u\n", mode);
+	LOGPTS(ts, DE1D, LOGL_INFO, "Starting in mode %s\n", get_value_string(e1_ts_mode_names, mode));
 
-	ret = socketpair(AF_UNIX, SOCK_SEQPACKET, 0, sd);
+	switch (mode) {
+	case E1_TS_MODE_HDLCFCS:
+		sock_type = SOCK_SEQPACKET;
+		break;
+	case E1_TS_MODE_RAW:
+		sock_type = SOCK_STREAM;
+		_e1d_ts_raw_buf_realloc(ts, bufsize);
+		break;
+	default:
+		return -EINVAL;
+	}
+
+	ret = socketpair(AF_UNIX, sock_type, 0, sd);
 	if (ret < 0)
 		return ret;
 
@@ -121,8 +141,8 @@ _e1d_ts_start(struct e1_ts *ts, enum e1_ts_mode mode)
 	ts->mode = mode;
 
 	if (mode == E1_TS_MODE_HDLCFCS) {
-		osmo_isdnhdlc_out_init(&ts->hdlc_tx, OSMO_HDLC_F_BITREVERSE);
-		osmo_isdnhdlc_rcv_init(&ts->hdlc_rx, OSMO_HDLC_F_BITREVERSE);
+		osmo_isdnhdlc_out_init(&ts->hdlc.tx, OSMO_HDLC_F_BITREVERSE);
+		osmo_isdnhdlc_rcv_init(&ts->hdlc.rx, OSMO_HDLC_F_BITREVERSE);
 	}
 
 	int flags = fcntl(ts->fd, F_GETFL);
@@ -254,6 +274,53 @@ _e1d_ctl_ts_query(void *data, struct msgb *msgb, struct msgb *rmsgb, int *rfd)
 	return 0;
 }
 
+static int
+_e1d_ctl_line_config(void *data, struct msgb *msgb, struct msgb *rmsgb, int *rfd)
+{
+	struct e1_daemon *e1d = (struct e1_daemon *)data;
+	struct osmo_e1dp_msg_hdr *hdr = msgb_l1(msgb);
+	struct osmo_e1dp_line_config *cfg = msgb_l2(msgb);
+	struct osmo_e1dp_line_info *info;
+	struct e1_intf *intf = NULL;
+	struct e1_line *line = NULL;
+
+	/* Process query and find timeslot */
+	intf = e1d_find_intf(e1d, hdr->intf);
+	if (!intf)
+		return 0;
+
+	line = e1_intf_find_line(intf, hdr->line);
+	if (!line)
+		return 0;
+
+	LOGPLI(line, DE1D, LOGL_NOTICE, "Setting line mode from %s to %s\n",
+		get_value_string(e1_line_mode_names, line->mode),
+		get_value_string(osmo_e1dp_line_mode_names, cfg->mode));
+	/* Select mode */
+	switch (cfg->mode) {
+	case E1DP_LMODE_CHANNELIZED:
+		line->mode = E1_LINE_MODE_CHANNELIZED;
+		break;
+	case E1DP_LMODE_SUPERCHANNEL:
+		line->mode = E1_LINE_MODE_SUPERCHANNEL;
+		break;
+	default:
+		return 0;
+	}
+
+	/* Allocate response */
+	rmsgb->l2h = msgb_put(rmsgb, sizeof(struct osmo_e1dp_line_info));
+	info = msgb_l2(rmsgb);
+
+	memset(info, 0x00, sizeof(struct osmo_e1dp_line_info));
+
+	/* Fill reponse */
+	_e1d_fill_line_info(info, line);
+
+	return 0;
+}
+
+
 static int
 _e1d_ctl_ts_open(void *data, struct msgb *msgb, struct msgb *rmsgb, int *rfd)
 {
@@ -269,16 +336,22 @@ _e1d_ctl_ts_open(void *data, struct msgb *msgb, struct msgb *rmsgb, int *rfd)
 
 	/* Process query and find timeslot */
 	intf = e1d_find_intf(e1d, hdr->intf);
-	if (!intf)
+	if (!intf) {
+		LOGP(DE1D, LOGL_NOTICE, "Client request for non-existant Interface %u\n", hdr->intf);
 		return 0;
+	}
 
 	line = e1_intf_find_line(intf, hdr->line);
-	if (!line)
+	if (!line) {
+		LOGPIF(intf, DE1D, LOGL_NOTICE, "Client request for non-existant line %u\n", hdr->line);
 		return 0;
+	}
 
 	ts = _e1d_get_ts(line, hdr->ts);
-	if (!ts)
+	if (!ts) {
+		LOGPLI(line, DE1D, LOGL_NOTICE, "Client request for non-existant ts %u\n", hdr->ts);
 		return 0;
+	}
 
 	/* Select mode */
 	switch (cfg->mode) {
@@ -289,16 +362,32 @@ _e1d_ctl_ts_open(void *data, struct msgb *msgb, struct msgb *rmsgb, int *rfd)
 		mode = E1_TS_MODE_HDLCFCS;
 		break;
 	default:
+		LOGPTS(ts, DE1D, LOGL_NOTICE, "Client request for unknown mode %u\n", cfg->mode);
 		return 0;
 	}
 
-	/* If already open, close previous */
-	e1_ts_stop(ts);
+	if (cfg->read_bufsize == 0) {
+		LOGPTS(ts, DE1D, LOGL_NOTICE, "Client request for invalid bufsize %u\n", cfg->read_bufsize);
+		return 0;
+	}
+
+	if (ts->fd >= 0) {
+		/* we only force-reopen a busy timeslot if the flag is set */
+		if (cfg->flags & E1DP_TS_OPEN_F_FORCE) {
+			e1_ts_stop(ts);
+		} else {
+			LOGPTS(ts, DE1D, LOGL_ERROR,
+				"Timeslot already open, rejecting re-open without F_FORCE\n");
+			return 0;
+		}
+	}
 
 	/* Init */
-	ret = _e1d_ts_start(ts, mode);
-	if (ret < 0)
+	ret = _e1d_ts_start(ts, mode, cfg->read_bufsize);
+	if (ret < 0) {
+		LOGPTS(ts, DE1D, LOGL_ERROR, "Unable to start timeslot: %d\n", ret);
 		return ret;
+	}
 
 	*rfd = ret;
 
@@ -336,6 +425,12 @@ struct osmo_e1dp_server_handler e1d_ctl_handlers[] = {
 		.payload_len = 0,
 		.fn = _e1d_ctl_ts_query,
 	},
+	{
+		.type = E1DP_CMD_LINE_CONFIG,
+		.flags = E1DP_SF_INTF_REQ | E1DP_SF_LINE_REQ,
+		.payload_len = sizeof(struct osmo_e1dp_line_config),
+		.fn = _e1d_ctl_line_config,
+	},
 	{
 		.type = E1DP_CMD_TS_OPEN,
 		.flags = E1DP_SF_INTF_REQ | E1DP_SF_LINE_REQ | E1DP_SF_TS_REQ,
